// © 2023 John Breaux
//! A bare-bones toy assembler for the TI MSP430, for use in MicroCorruption
//!
//! This project aims to assemble any valid msp430 instructions, while being lenient about the
//! syntax. After all, a real-world parser is going to face all kinds of malformed input, and it
//! would be nice to support that kind of input (or, if it's completely unsalvageable, provide a
//! useful message to the author.)
//!
//! The [`Parser`](preamble::Parser) will ignore whitespace, excluding newlines,
//! unless syntactically relevant. It will also discard comma-separators between operands of a
//! two-operand instruction.
//!
//! It returns an AST structured as follows
//! ```text
//! Root
//! ├─ Line
//! │  └─ Empty
//! ├─ Line
//! │  └─ Comment
//! ├─ Line
//! │  └─ Directive                 // Pre- or Post-processor directive
//! ├─ Linel
//! │  └─ Label                     // Label definition
//! ├─ Line
//! │  └─ Instruction
//! │     ├─ Opcode
//! │     └─ Encoding::Single
//! │        ├─ Width
//! │        └─ PrimaryOperand
//! │           ├─ Identifier       // Label, for relative-addressed data/code
//! │           ├─ Register         // Direct, indexed, indirect or indirect-post-increment register.
//! │           └─ Number           // Index, absolute address or immediate value.
//! ├─ Line
//! │  └─ Instruction
//! │     ├─ Opcode
//! │     └─ Encoding::Double
//! │        ├─ Width
//! │        ├─ PrimaryOperand
//! │           ├─ Identifier       // Label, for relative-addressed data/code
//! │        │  ├─ Register         // Direct, indexed, indirect or indirect-post-increment register.
//! │        │  └─ Number           // Index, absolute address or immediate value.
//! │        └─ SecondaryOperand
//! │           ├─ Identifier       // Label, for relative-addressed data/code
//! │           ├─ Register         // Direct or indexed register
//! │           └─ Number           // Index or absolute address
//! ├─ Line
//! │  └─ Instruction
//! │     ├─ Opcode
//! │     └─ Encoding::Jump
//! │        └─ JumpTarget
//! │           ├─ Identifier       // Label
//! │           └─ Number           // Even, PC-relative offset in range (-1024..=1022)
//! └─ Line
//!    └─ EndOfFile
//! ```

pub mod preamble {
    //! Common imports for msp430-asm
    use super::*;
    pub use assembler::Assembler;
    pub use error::Error;
    pub use lexer::{
        context::Context,
        token::{Token, Type},
        token_stream::TokenStream,
        Tokenizer,
    };
    pub use parser::Parser;
}

use preamble::*;
pub mod error;

pub mod assembler;
pub mod lexer;
pub mod parser;
