// © 2023 John Breaux
//! A [`Directive`] issues commands directly to the [`Tokenizer`](crate::Tokenizer) and
//! [Linker](crate::Linker)

use std::path::PathBuf;

use super::*;
use crate::lexer::token::OwnedToken;

// TODO: Parse each kind of *postprocessor* directive into an AST node
//  - .org 8000:                Directive::Org { base: Number }
//  - .define ident tt...       Directive::Define { } ; should this be in the AST? How do I put this
//    in the AST?
//  - .include "<filename>"     Directive::Include { Root } ; should this include an entire AST in
//    the AST?
//  - .word 8000                Directive::Word(Number)
//  - .words dead beef          Directive::Words(Vec<u16>|Vec<Number>)
//  - .byte ff                  Directive::Byte(Number)
//  - .bytes de, ad, be, ef     Directive::Bytes(Vec<u8>)
//  - .string "string"          Directive::String(String)
//  - .ascii "string"           Directive::Ascii(Vec<u8>)

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Directive {
    Org(Number),
    Define(Vec<OwnedToken>),
    Include(Root),
    Byte(Number),
    Bytes(Vec<Number>),
    Word(Number),
    Words(Vec<Number>),
    String(String),
    Strings(Vec<String>),
}

impl Directive {}

impl Parsable for Directive {
    fn parse<'text, T>(p: &Parser, stream: &mut T) -> Result<Self, ParseError>
    where T: TokenStream<'text> {
        let d = stream.expect(Type::Directive)?;
        // match on the directive
        Ok(match d.lexeme() {
            ".org" => Self::Org(Number::parse(p, stream)?),
            ".define" => {
                let mut tokens = vec![];
                loop {
                    match stream.peek().variant() {
                        Type::Endl | Type::EndOfFile => break,
                        _ => tokens.push(stream.next().unwrap_or_default().into()),
                    }
                }
                Self::Define(tokens)
            }
            ".include" => {
                // Try to get path
                Self::Include(Parser::default().parse_file(&PathBuf::parse(p, stream)?)?)
            }
            ".byte" => Self::Byte(Number::parse(p, stream)?),
            ".bytes" => Self::Bytes(Vec::<Number>::parse(p, stream)?),
            ".word" => Self::Word(Number::parse(p, stream)?),
            ".words" => Self::Words(Vec::<Number>::parse(p, stream)?),
            ".string" => Self::String(String::parse(p, stream)?),
            ".strings" => Self::Strings(Vec::<String>::parse(p, stream)?),
            e => Err(ParseError::UnrecognizedDirective(e.into()))?,
        })
    }
}

impl Display for Directive {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Directive::Org(num) => write!(f, ".org {num}"),
            Directive::Define(rep) => {
                write!(f, ".define")?;
                for t in rep {
                    write!(f, " {t}")?;
                }
                Ok(())
            }
            Directive::Include(r) => Display::fmt(r, f),
            Directive::Byte(num) => write!(f, ".org {num}"),
            Directive::Bytes(v) => write!(f, ".bytes {v:?}"),
            Directive::Word(num) => write!(f, ".org {num}"),
            Directive::Words(v) => write!(f, ".bytes {v:?}"),
            Directive::String(s) => write!(f, ".string \"{s}\""),
            Directive::Strings(s) => write!(f, ".string \"{s:?}\""),
        }
    }
}
