// © 2023 John Breaux
//! A [`Number`] represents a 16-bit signed or unsigned word
use super::*;

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Number(isize, u32); // (value, radix)

impl Parsable for Number {
    // A number is:
    // [Minus|Plus]? RadixMarker[Hex|Dec|Oct|Bin]? Number
    fn parse<'text, T>(p: &Parser, stream: &mut T) -> Result<Self, ParseError>
    where T: TokenStream<'text> {
        use Type as Ty;
        // The number is negative when it begins with a Minus, but Plus is also acceptable.
        let negative = stream.expect_any_of([Ty::Minus, Ty::Plus]).map_or(false, |t| t.is_variant(Ty::Minus));
        let radix = match stream
            .expect_any_of([Ty::RadixMarkerHex, Ty::RadixMarkerDec, Ty::RadixMarkerOct, Ty::RadixMarkerBin])
            .ok()
            .map(|t| t.variant())
        {
            Some(Ty::RadixMarkerHex) => 16,
            Some(Ty::RadixMarkerDec) => 10,
            Some(Ty::RadixMarkerOct) => 8,
            Some(Ty::RadixMarkerBin) => 2,
            _ => p.radix,
        };
        let number = stream.expect(Ty::Number)?;
        // TODO: Reintroduce error context
        let number = isize::from_str_radix(number.lexeme(), radix)
            .map_err(|_| ParseError::UnexpectedDigits(number.lexeme().into(), radix))?
            * if negative { -1 } else { 1 };
        // Ensure number fits within a *signed or unsigned* 16-bit int (it will be truncated to fit)
        Ok(Self(
            if (-0x8000..0x10000).contains(&number) { number } else { Err(ParseError::NumberTooWide(number))? },
            radix,
        ))
    }
}

impl From<isize> for Number {
    fn from(value: isize) -> Self { Self(value, 16) }
}

impl From<Number> for isize {
    fn from(value: Number) -> Self { value.0 as Self }
}

impl From<u16> for Number {
    fn from(value: u16) -> Self { Self(value as isize, 16) }
}

impl From<Number> for u16 {
    fn from(value: Number) -> Self { value.0 as Self }
}

impl std::ops::Sub<isize> for Number {
    type Output = Self;
    fn sub(mut self, rhs: isize) -> Self::Output {
        self.0 -= rhs;
        self
    }
}

impl std::ops::Shr<usize> for Number {
    type Output = Self;
    fn shr(mut self, rhs: usize) -> Self::Output {
        self.0 >>= rhs;
        self
    }
}

impl std::fmt::Display for Number {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.1 {
            2 => std::fmt::Binary::fmt(&self.0, f),
            8 => std::fmt::Octal::fmt(&self.0, f),
            16 => std::fmt::LowerHex::fmt(&self.0, f),
            _ => std::fmt::Display::fmt(&self.0, f),
        }
    }
}
