// © 2023 John Breaux
//! An [`Opcode`] encodes an msp430 operation
use super::*;

use std::str::FromStr;

/// Opcode from the [MSPGCC Manual][1]
///
/// Calling [`resolve()`](Opcode::resolve()) will emit an [EncodingParser] which will
/// extract from a [TokenStream] only the required arguments for that call.
///
/// [1]: https://mspgcc.sourceforge.net/manual/x223.html
#[allow(clippy::identity_op)]
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Opcode {
    // "Emulated" opcodes
    Nop,
    Pop,
    Br,
    Ret,
    Clrc,
    Setc,
    Clrz,
    Setz,
    Clrn,
    Setn,
    Dint,
    Eint,
    Rla,
    Rlc,
    Inv,
    Clr,
    Tst,
    Dec,
    Decd,
    Inc,
    Incd,
    Adc,
    Dadc,
    Sbc,
    // Single
    Rrc  = 0x1000 | 0 << 7,
    Swpb = 0x1000 | 1 << 7,
    Rra  = 0x1000 | 2 << 7,
    Sxt  = 0x1000 | 3 << 7,
    Push = 0x1000 | 4 << 7,
    Call = 0x1000 | 5 << 7,
    Reti = 0x1000 | 6 << 7,
    // Jump
    Jnz  = 0x2000 | 0 << 10,
    Jz   = 0x2000 | 1 << 10,
    Jnc  = 0x2000 | 2 << 10,
    Jc   = 0x2000 | 3 << 10,
    Jn   = 0x2000 | 4 << 10,
    Jge  = 0x2000 | 5 << 10,
    Jl   = 0x2000 | 6 << 10,
    Jmp  = 0x2000 | 7 << 10,
    // Double
    Mov  = 0x4000,
    Add  = 0x5000,
    Addc = 0x6000,
    Subc = 0x7000,
    Sub  = 0x8000,
    Cmp  = 0x9000,
    Dadd = 0xa000,
    Bit  = 0xb000,
    Bic  = 0xc000,
    Bis  = 0xd000,
    Xor  = 0xe000,
    And  = 0xf000,
}

impl Opcode {
    /// Resolve an Opcode into an [Opcode] and an [EncodingParser]
    pub fn resolve(self) -> (Opcode, EncodingParser) {
        use super::Encoding as Enc;
        use Register as Reg;
        use {PrimaryOperand as Src, SecondaryOperand as Dst};
        match self {
            Self::Rrc | Self::Rra | Self::Push => (self, Enc::single().end()),
            // these instructions do not take a width specifier (though they may still behave properly)
            Self::Swpb | Self::Sxt | Self::Call => (self, Enc::single().width(false).end()),
            // `reti` does not take any operands.
            Self::Reti => (self, Enc::single().operand(Src::Direct(Reg::pc)).end()),
            Self::Jnz | Self::Jz | Self::Jnc | Self::Jc | Self::Jn | Self::Jge | Self::Jl | Self::Jmp => {
                (self, Enc::jump().end())
            }
            Self::Mov
            | Self::Add
            | Self::Addc
            | Self::Subc
            | Self::Sub
            | Self::Cmp
            | Self::Dadd
            | Self::Bit
            | Self::Bic
            | Self::Bis
            | Self::Xor
            | Self::And => (self, Enc::double().end()),
            Self::Nop => (Self::Mov, Enc::double().src(Src::Zero).dst(Dst::Zero).end()),
            Self::Pop => (Self::Mov, Enc::double().src(Src::PostInc(Reg::sp)).end()),
            Self::Br => (Self::Mov, Enc::double().dst(Dst::Direct(Reg::pc)).end()),
            Self::Ret => (Self::Mov, Enc::double().src(Src::PostInc(Reg::sp)).dst(Dst::Direct(Reg::pc)).end()),
            Self::Clrc => (Self::Bic, Enc::double().src(Src::One).dst(Dst::Direct(Reg::sr)).end()),
            Self::Setc => (Self::Bis, Enc::double().src(Src::One).dst(Dst::Direct(Reg::sr)).end()),
            Self::Clrz => (Self::Bic, Enc::double().src(Src::Two).dst(Dst::Direct(Reg::sr)).end()),
            Self::Setz => (Self::Bis, Enc::double().src(Src::Two).dst(Dst::Direct(Reg::sr)).end()),
            Self::Clrn => (Self::Bic, Enc::double().src(Src::Four).dst(Dst::Direct(Reg::sr)).end()),
            Self::Setn => (Self::Bis, Enc::double().src(Src::Four).dst(Dst::Direct(Reg::sr)).end()),
            Self::Dint => (Self::Bic, Enc::double().src(Src::Eight).dst(Dst::Direct(Reg::sr)).end()),
            Self::Eint => (Self::Bis, Enc::double().src(Src::Eight).dst(Dst::Direct(Reg::sr)).end()),
            Self::Rla => (Self::Add, Enc::reflexive().end()),
            Self::Rlc => (Self::Addc, Enc::reflexive().end()),
            Self::Inv => (Self::Xor, Enc::double().src(Src::MinusOne).end()),
            Self::Clr => (Self::Mov, Enc::double().src(Src::Zero).end()),
            Self::Tst => (Self::Cmp, Enc::double().src(Src::Zero).end()),
            Self::Dec => (Self::Sub, Enc::double().src(Src::One).end()),
            Self::Decd => (Self::Sub, Enc::double().src(Src::Two).end()),
            Self::Inc => (Self::Add, Enc::double().src(Src::One).end()),
            Self::Incd => (Self::Add, Enc::double().src(Src::Two).end()),
            Self::Adc => (Self::Addc, Enc::double().src(Src::Zero).end()),
            Self::Dadc => (Self::Dadd, Enc::double().src(Src::Zero).end()),
            Self::Sbc => (Self::Subc, Enc::double().src(Src::Zero).end()),
        }
    }
}

impl Parsable for Opcode {
    fn parse<'text, T>(_: &Parser, stream: &mut T) -> Result<Self, ParseError>
    where T: TokenStream<'text> {
        // TODO: Reintroduce error context
        stream.expect(Type::Insn)?.parse()
    }
}

impl FromStr for Opcode {
    type Err = ParseError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        //TODO: Reduce allocations here?
        let s = s.to_ascii_lowercase();
        Ok(match s.as_str() {
            "rrc" => Self::Rrc,
            "swpb" => Self::Swpb,
            "rra" => Self::Rra,
            "sxt" => Self::Sxt,
            "push" => Self::Push,
            "call" => Self::Call,
            "reti" => Self::Reti,

            "jne" | "jnz" => Self::Jnz,
            "jeq" | "jz" => Self::Jz,
            "jnc" | "jlo" => Self::Jnc,
            "jc" | "jhs" => Self::Jc,
            "jn" => Self::Jn,
            "jge" => Self::Jge,
            "jl" => Self::Jl,
            "jmp" => Self::Jmp,

            "mov" => Self::Mov,
            "add" => Self::Add,
            "addc" => Self::Addc,
            "subc" => Self::Subc,
            "sub" => Self::Sub,
            "cmp" => Self::Cmp,
            "dadd" => Self::Dadd,
            "bit" => Self::Bit,
            "bic" => Self::Bic,
            "bis" => Self::Bis,
            "xor" => Self::Xor,
            "and" => Self::And,

            "nop" => Self::Nop,
            "pop" => Self::Pop,
            "br" => Self::Br,
            "ret" => Self::Ret,
            "clrc" => Self::Clrc,
            "setc" => Self::Setc,
            "clrz" => Self::Clrz,
            "setz" => Self::Setz,
            "clrn" => Self::Clrn,
            "setn" => Self::Setn,
            "dint" => Self::Dint,
            "eint" => Self::Eint,
            "rla" => Self::Rla,
            "rlc" => Self::Rlc,
            "inv" => Self::Inv,
            "clr" => Self::Clr,
            "tst" => Self::Tst,
            "dec" => Self::Dec,
            "decd" => Self::Decd,
            "inc" => Self::Inc,
            "incd" => Self::Incd,
            "adc" => Self::Adc,
            "dadc" => Self::Dadc,
            "sbc" => Self::Sbc,
            _ => Err(ParseError::UnrecognizedOpcode(s))?,
        })
    }
}

impl Display for Opcode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Nop => "nop",
                Self::Pop => "pop",
                Self::Br => "br",
                Self::Ret => "ret",
                Self::Clrc => "clrc",
                Self::Setc => "setc",
                Self::Clrz => "clrz",
                Self::Setz => "setz",
                Self::Clrn => "clrn",
                Self::Setn => "setn",
                Self::Dint => "dint",
                Self::Eint => "eint",
                Self::Rla => "rla",
                Self::Rlc => "rlc",
                Self::Inv => "inv",
                Self::Clr => "clr",
                Self::Tst => "tst",
                Self::Dec => "dec",
                Self::Decd => "decd",
                Self::Inc => "inc",
                Self::Incd => "incd",
                Self::Adc => "adc",
                Self::Dadc => "dadc",
                Self::Sbc => "sbc",
                Self::Rrc => "rrc",
                Self::Swpb => "swpb",
                Self::Rra => "rra",
                Self::Sxt => "sxt",
                Self::Push => "push",
                Self::Call => "call",
                Self::Reti => "reti",
                Self::Jnz => "jnz",
                Self::Jz => "jz",
                Self::Jnc => "jnc",
                Self::Jc => "jc",
                Self::Jn => "jn",
                Self::Jge => "jge",
                Self::Jl => "jl",
                Self::Jmp => "jmp",
                Self::Mov => "mov",
                Self::Add => "add",
                Self::Addc => "addc",
                Self::Subc => "subc",
                Self::Sub => "sub",
                Self::Cmp => "cmp",
                Self::Dadd => "dadd",
                Self::Bit => "bit",
                Self::Bic => "bic",
                Self::Bis => "bis",
                Self::Xor => "xor",
                Self::And => "and",
            }
        )
    }
}
